#!/usr/libexec/platform-python
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import os, string, sys, types
from atb.ctpemu import check_emubatchcmd
from auxfuns import parseline,printError,getfile,isinargs
from atb.bcmasktxt import BCmask
from atb.bcmask import symname

#!/usr/bin/python3
#!/usr/bin/python3
LTUS=None

def getN4tdimux(ltuboardname):
  """rc: 1,2, or 4  (det1 det2 det3) for tdimux/global in LTU
None: no config available
""" 
  cfgdir= os.getenv("CTP3_DBCTP")
  if cfgdir==None: return None
  netcfg= getfile(os.path.join(cfgdir,"ctp.net"), errprint=False)
  if netcfg==None: return None
  linen= -1
  for ixl in range(len(netcfg)):
    #atb_64
    #1 atb_37 atb_38 atb_39
    #2 atb_40 atb_41 atb_43
    #3 atb_51 atb_50 atb_44
    #4 atb_52 # atb_53
    #5 atb_45 atb_46
    #6 atb_47 atb_48 atb_49
    line= parseline(netcfg[ixl])
    if line==None: continue
    if linen==-1:    # 1st line of netcfg
      linen= 0
      ctpbname= line[0]
      continue
    if not (line[0] in ('1','2','3','4','5','6')): continue   # gbtX...
    oltn= int(line[0])
    locemun= 1   # 1,2 or 3
    for ixbn in range(len(line[1:])):
      ltubname= line[ixbn+1]
      if ltubname==ctpbname:
        print("getN4tdimux MONONU for OLT%d, ignored"%oltn)
        continue
      #if ltu==None:
      #  print("Unknown board:", ltubname, " in ctp.net")
      #  continue
      emun= (oltn-1)*3 + locemun
      if ltubname==ltuboardname:
        # emun: 1..18  %3 -> 1,2,3  but bit for mux -> 1 2 4
        det124= ((emun-1) % 3)+1
        if det124==3: det124= 4
        return det124
      locemun= locemun+1
  return None
def findEmuN(ltuname):
  """ check if ltuname has a ttcpon connection to the ctp
rc: None
    1..18 in case connected
"""
  if ltuname in LTUS.ltus:
    if LTUS.ltus[ltuname]== None: return None
    return LTUS.ltus[ltuname].emun
  printError("Internal: findEmuN(%s)"%ltuname)
  return 0
class Ltu(object):
  def __init__(self, name, bname):
    self.name= name
    self.bname= bname
    self.emun= None   # 1..18, None: not present in ctp.net, i.e. ltu not wired/wired
    self.ltustatus= None # available, active (std/global?), locked (by?)
class Ltus(object):
  def __init__(self):
    """
I:
1. ctpboards.cfg
2. ctp.net
outcome:
self.ltus: {} of Ltu refs, key: ltu name
self.ltun: [] of Ltu refs ordered by emu number
self.ctpbname the name of ctp board:
     -the ctp-line in ctpboards.net is read, but ctpbname overwritten by
     -value in 1st line in ctp.net (if ctp.net exists)
"""
    #print("partdef.LTUS constructor")
    self.ctpbname= None   # ctpboards.cfg overwritten by ctp.net
    self.ltus={ "trd":None, "hmp":None, "phs":None, "emc":None, "cpv":None,
      "fv0":None, "ft0":None, "fdd":None, "its":None, "tof":None,
      "mft":None, "mid":None, "mch":None, "tpc":None, "zdc":None,
      "daq":None}
    cfgdir= os.getenv("CTP3_DBCTP")
    if cfgdir==None: return
    netcfg= getfile(os.path.join(cfgdir,"ctpboards.cfg"), errprint=False)
    if netcfg==None: return
    #print("$CTP3_DBCTP dir processing...")
    for ixl in range(len(netcfg)):
      #trd 45 c26t 101 102 lvpecl
      #spare -
      line= parseline(netcfg[ixl])
      if line==None: continue
      if len(line)<2: continue
      if line[1]=='0': continue
      name= line[0]
      if name in self.ltus:
        self.ltus[name]= Ltu(name, "atb_"+line[1])
      elif name == "ctp":
        self.ctpbname= "atb_"+line[1]
      else:
        print("strange LTU name in ctpboards.cfg:{}".format(name))
    #
    netcfg= getfile(os.path.join(cfgdir,"ctp.net"))
    if netcfg==None: return
    linen= -1
    for ixl in range(len(netcfg)):
      #atb_64
      #1 atb_37 atb_38 atb_39
      #2 atb_40 atb_41 atb_43
      #3 atb_51 atb_50 atb_44
      #4 atb_52 # atb_53
      #5 atb_45 atb_46
      #6 atb_47 atb_48 atb_49
      line= parseline(netcfg[ixl])
      if line==None: continue
      if linen==-1:    # 1st line of netcfg
        linen= 0
        self.ctpbname= line[0]   # higher priority if ctp.net exists
        continue
      if not (line[0] in ('1','2','3','4','5','6')): continue   # ignore gbtX...
      oltn= int(line[0])
      locemun= 1   # 1,2 or 3
      for ixbn in range(len(line[1:])):
        ltubname= line[ixbn+1]
        if ltubname==self.ctpbname:
          #print("MONONU for OLT%d, ignored"%oltn)
          continue
        ltu= self.findBN(ltubname)
        if ltu==None:
          printError("Unknown board: {} in ctp.net".format(ltubname))
          continue
        ltu.emun= (oltn-1)*3 + locemun
        locemun= locemun+1
    #self.prt()
  def findDN(self, detname, wired=False):
    """ check if detname is:
wired  check
------------
False  valid:   i.e. fv0, tpc,hmp,... is a valid name
True   wired:   valid + wired with ctp  (stronger test)
rc: bname       -ok, ltu wired to the ctp
    atb_detname -board not assigned (missing in ctpboards.cfg)
    None        -detname not defined in Alice
"""
    if detname in self.ltus:
      # detname is valid, i.e. exists in ctpboards.cfg
      if self.ltus[detname]==None:
        # but board not wired/connected to the ctp
        if wired:
          return None
        return "atb_"+detname   # valid, board not assigned, i.e. cannot be wired
      else:
        if wired:
          if self.ltus[detname].emun==None:
            return None   # not wired i.e. not present in ctp.net
        return self.ltus[detname].bname
    return None
  def findBN(self, bname):   # rc: Ltu ref or None
    for ltuname in self.ltus:
      if self.ltus[ltuname]==None: continue
      if self.ltus[ltuname].bname==bname: return self.ltus[ltuname]
    return None
  def prt(self):
    line= ""
    for ltun in self.ltus:
      ltu= self.ltus[ltun]
      if ltu:
        line= line+"%s %s:%s, "%(ltu.name, ltu.bname, str(ltu.emun))
    print("Ltus:",line)
class Resources(object):
  """ object connected with a Partition object or Ctp object.
Ctp.Resources = sum of all Partition objects resources, sharing 
taken into account:
clusters,ltus: cannot be shared between 2 partitions (private resource)
clusters: probably no need to watch on this (# of clusters = # of ltus)
classes: detto (each class has a dedicated cluster, it cannot drive
         2 clusters
bdg1/2 rng1/2: can be shared if value is equal (shared resource)

Resources object is result of:
- rsrcs= Partition.getResources()
- rsrcs= rsrcs1 + rsrcs2 + ...
"""
  def __init__(self, ltus=None):
    """ ltus: list of dets
"""
    #self.classes= 0
    #self.clusters= 0
    self.ltus= {}       # {"hmp":N,...}  N>1: overload
    if ltus!=None:
      for ltu in ltus:
        self.ltus[ltu]= 1
    #self.bdgens= {}    # {"1":value, "2":value,... }
    #self.rngens= {}    # {"1":value,... }
    # X where X is 1,2,3,... (if X>2: an overload of resources)
    # allocation of X to bc1 or bc2 is done in Ctp
    #self.masks= 0
    #self.masknames= {}
  def __add__(self, other):
    rc= Resources()
    # ltus:
    rc.ltus= self.ltus.copy()
    rc.ltus.update(other.ltus)
    # rc.ltus: keys from both, if were in both other was taken
    for ltu in rc.ltus:   # for all
      # 1: self 2:other 12:in both
      if ltu in self.ltus:
        if ltu in other.ltus:   # 12
          rc.ltus[ltu]= self.ltus[ltu] + other.ltus[ltu]
        else:   # 1
          pass   # (copied already in copy()
      else:     # 2 
        pass    # was copied in update()
    #print("Resources.__add_ rc:{}".format(rc))
    return rc
  def check(self, allowedltus):
    """
self.allowedltus!=None: check also the dict. of allowed ltus (ctpd knows 
which are involved in stdalone run/not ready).
Note: this mechanism not used for global runs, these are treated directly
in Resources object for ctp 
"""
    #if len(self.ltus)>18:
    #print("Resources.check:{}".format(self.ltus))
    for ltu in self.ltus:
      if (allowedltus !=None) and (ltu not in allowedltus):
        print("LTU {} not available, i.e. already in global run, or not wired".format(ltu))
        return False
      if self.ltus[ltu]>1:
        print("LTU {} already assigned to another partition".format(ltu))
        return False
    return True
  def prt(self):
    print("LTUS:",self.ltus)
    
class BCM(object):
  def __init__(self, name):
    self.maskdef= None
    self.name= name
  def setMaskDef(self, maskdef):
    # check and set mask definition
    msk= BCmask()   # check syntax
    msk.update(maskdef)
    if msk.errmsg=="":
      #self.maskdef= maskdef
      self.maskdef= msk.bcmstr
    else:
      printError(msk.errmsg)
class LTG(object):
  def __init__(self,ltgline, partinst):
    #print("LTG:", ltgline)
    self.partinst= partinst
    self.detname= ltgline[0]
    self.loadcmds= []    # list of cmds to load LTG
    n= findEmuN(self.detname)
    if n==None:
      self.emuname= None
      self.partinst.regError("Unknown detector {} in: LTG {}".format(self.detname,ltgline))
    else:
      self.emuname= "emu%d"%n
    # emuname= findEmuN()
  def loadcmdsAppend(self, cmd):
    emsg= check_emubatchcmd(cmd)  # emuX par1 par2
    if emsg!="":
      self.partinst.regError(emsg+" "+" ".join(cmd))
    else:
      self.loadcmds.append(cmd)   # emuX par1 par2 ...
    return emsg
class Cluster(object):
  def __init__(self, cluline, part):
    """ cluline: [cluster_name, det1, det2,...] , detectors already applied in proccluster
"""
    global LTUS
    #print("Cluster:",cluline)
    self.cluname= cluline[0]
    self.detnames= []
    self.classes= []
    grundets= 0   # 18 bits
    for ixd in range(1,len(cluline)):
      detname= cluline[ixd]
      #print detname,ixd
      if (detname in LTUS.ltus) and (LTUS.ltus[detname]!=None):
        emun= LTUS.ltus[detname].emun
        if (part.allowedltus !=None) and (detname not in part.allowedltus):
          part.regError("Cluster %s: detector:%s not available (in stdalone run?)"%(self.cluname,detname))
        elif emun==None:
          part.regError("Cluster %s: detector:%s not connected"%(self.cluname,detname))
        else:
          grundets= grundets | (1<<(emun-1))
          self.detnames.append(detname)
        #print "Cluster det:", detname, hex(grundets)
      else:
        part.regError("Cluster %s: unknown detector:%s"%(self.cluname,detname))
    if grundets==0: grundets= None
    self.grundets= grundets
class Partition(object):
  STOPPED=5    # stopped, being deleted
  LOADED=4     # partition def. read in from db, not started
  LOADEDHW=2   # LOADED + loaded in hw
  PAUSED= 1
  RUNNING=0    # active (see ctpd.proto RunReply)
  def __init__(self, pname, runn, detectors="", allowedltus=None):   
    """
pname: filename[.pd] (in $CTP3_DBCTP/pd/) or
       part. definition (list of strings/lines or long string containing \n)
detectors: detectors (space separated) to be used in the run (subset 
       of all pname detectors). Has to be applied before allowedltus check.
allowedltus: list of available ltus (wired + in stdalone) -error
       in case of attempt to use it

Out: self.status LOADED -ok
                 none   -problem, see self.errs+self.lalerrs
"""
    self.clusters= []
    self.ltgs= {}   # key: detname
    self.masks= {}
    self.status= None   #  will become LOADED after successful load 
    self.errs= 0    # number of Errors (all in self.allerrs)
    self.allerrs= ""
    self.runn= runn
    if type(pname)==types.ListType:
      netcfg= pname
      pname= "run{}".format(runn)   # direct list of strings
    elif len(pname)>24:
      netcfg= pname.split('\n')
      pname= "run{}".format(runn)   # direct string with NLs
    else:
      #if len(pname)>24:
      #  self.regError("Too long part. name %24s...".format(pname))
      #  return
      if pname[-3:]==".pd": pname= pname[:-3]
      cfgdir= os.getenv("CTP3_DBCTP")
      fname= os.path.join(cfgdir,"pd",pname+".pd")
      netcfg= getfile(fname)
      if netcfg==None:
        self.regError("{} does not exist.".format(fname))
        return
    self.pname= pname
    if detectors=="":
      detectors= None
    else:
      detectors= detectors.split()
    self.detectors= detectors
    self.allowedltus= allowedltus
    self.activesec= None   # ref to an active Cluster or LTG instance
    ignorelines= False
    for ixl in range(len(netcfg)):
      #cluster name det1 det2 ...
      self.line= parseline(netcfg[ixl])
      if self.line==None: continue
      if ignorelines:
        if self.keyw("bcm",self.line[0]) or self.keyw("ltg",self.line[0]) or\
          self.keyw("cluster",self.line[0]):
          ignorelines= False
        else:
          continue
      #
      # sections in fixed order:
      # None (common)
      # LTG
      # Cluster
      #
      # LTG: bcm,...   cluster ->finish LTG
      # todo: class
      if self.activesec== None:   # ====== common section
        # bcm ltg cluster
        if self.line[0]=='bcm':
          # todo: check self.masks if already in
          if len(self.line)==1:
            self.regError("bcm name missing in: {}".format(self.line))
          maskname= self.line[1]
          if not (maskname in symname):
            self.regError("Incorrect mask name: {}".format(maskname))   
          bcm= BCM(self.line[1])
          bcm.setMaskDef(" ".join(self.line[2:]))
          if bcm.maskdef!=None:
            self.masks[maskname]= bcm
          else:
            self.regError("Incorrect definition of {} mask in:{}".format(maskname,self.line))
        elif self.keyw('ltg', self.line[0]):
          if self.procltg()==None:   # err in LTG line or thrown out by detectors
            #break
            ignorelines=True
        elif self.line[0]=='cluster':
          if self.proccluster()==None: ignorelines=True
        else:
          self.regError("bcm, ltg or cluster key expected in line:%d"%ixl)
          #print("Line:",self.line)
      elif isinstance(self.activesec, LTG):       # ====== ltg section
        # bcm TOF enable
        # ltg detname
        # cluster cluname det1 det2 ...
        #print "LTG %s active. Line:"%self.activesec.detname,self.line
        #todo: apply detectors
        emuname= self.activesec.emuname
        if self.line[0]=='bcm':
          # special (should probably be removed) 'bcm TOF enable'
          # correct is 'emu bcm TOF enable'
          if len(self.line)==1:
            self.regError("bcm name missing for ltg "+dn)
          else:
            # "emuX bcm TOF enable"
            cmd= (emuname, "bcm", self.line[1],"enable")
            self.activesec.loadcmdsAppend(cmd)   # emuX...
        elif self.line[0]=='emu':
          cmd= (emuname,)+ tuple(self.line[1:])
          self.activesec.loadcmdsAppend(cmd)   # emuX...
        elif self.keyw('ltg', self.line[0]):
          if self.procltg()==None:
            #break
            ignorelines=True
        elif self.line[0]=='cluster':
          if self.proccluster()==None: ignorelines= True
        else:
          #self.regError("ltg %s definition wrong (only 'bcm emu' lines allowed for now)"%\
          #  self.activesec.detname)
          cmd= (emuname,)+ tuple(self.line[0:])
          self.activesec.loadcmdsAppend(cmd)   # emuX...
      elif isinstance(self.activesec, Cluster):   # ====== cluster
        if self.line[0]=='cluster':
          if self.proccluster()==None: ignorelines= True
        else:
          self.regError("cluster %s definition: body(i.e.classes) not allowed yet"%\
            self.activesec.cluname)
      else:
        # no classes yet...
        self.regError("Internal error (unknown activesec ?) in:"+self.line)
    if self.errs== 0:
      self.setStatus(self.LOADED)
  def getResources(self):
    #rc= Resources(ltus= self.ltgs.keys())   # generally, LTG not defined
    rc= Resources(ltus= self.getDets())
    return rc
  def keyw(self, resword, word):
    if resword==word.lower(): return True
    return False
  def regError(self, msg):
    self.allerrs= self.allerrs + msg + "\n"
    #printError(msg+". Partition:"+self.pname+", in line:\n"+string.join(self.line))
    self.errs+= 1
  def getDets18(self):
    allclus= 0
    for ix in range(len(self.clusters)):
      allclus= allclus | self.clusters[ix].grundets
    return allclus
  def getCluDets(self):   # rc: "clu1:d1,d2 clu2:d1,d2...:
    allclus= ""
    for ix in range(len(self.clusters)):
      allclus= allclus + self.clusters[ix].cluname + ":" +\
        ",".join(self.clusters[ix].detnames) + " "
    return allclus
  def getDets(self):   # rc: ["d1, "d2", "d3"...]
    alldets= []
    #for lname in self.ltgs: alldets.append(lname)  bad (not all ltus have LTG object)
    for ix in range(len(self.clusters)):
      for dn in self.clusters[ix].detnames:
        if dn not in alldets: alldets.append(dn)
    return alldets
  def procltg(self):
    """ rc: -None -> error reported in part.allerrs or
                  -> this LTG ignored becasue of detectors
            -new LTG instance -also added in self.ltgs and self.activesec set
"""
    if len(self.line)==1:
      self.regError("ltg name missing in:"+self.line)
      return None
    lname= self.line[1]
    if self.detectors and (lname not in self.detectors):
      print("Ignoring LTG {}".format(lname))
      return None
    if lname in self.ltgs:
      print("ltg %s already defined once, extending..."%lname)
      ltg= self.ltgs[lname]
    else:
      if LTUS.findDN(lname)==None:
        self.regError("Unknown LTG name {} in:{}".format(lname,' '.join(self.line)))
        return None
      # check if wired
      if findEmuN(lname)==None:
        self.regError("LTG {} not wired in line:{}".format(lname,' '.join(self.line)))
        return None
      ltg= LTG(self.line[1:], self)
      if ltg.emuname!=None:
        self.ltgs[lname]= ltg
      else:
        ltg= None
    self.activesec= ltg
    return ltg
  def proccluster(self):   # rc: see procltg
    if len(self.line)<2:
      self.regError("cluster cluster_name det1 [det2 ...] expected in line:"+\
        self.line)
      return None
    lname= self.line[1]
    clu= self.findclu(lname)
    if (clu != None):
      if len(self.line) == 2:
        print("cluster %s already defined once, extending..."%lname)
        # i.e. allowed: cluster anyname    (det list must be empty)
      else:
        self.regError("Detectors ignored in line {}".format(" ".join(self.line)))
    else:
      cludets= [lname]   # 1st item: cluster name
      for ix in range(len(self.line[2:])):
        dn= self.line[2+ix]
        if self.detectors:
          if dn not in self.detectors: continue
        cludets.append(dn)
      if len(cludets)==1:  
        print("Empty cluster %s, class defs ignored"%lname)
        return None
      clu= Cluster(cludets, self)
      if clu.grundets!=None:
        self.clusters.append(clu)
        self.activesec= clu   # current one
      else:
        clu= None
    return clu
  def findclu(self,cluname):
    for ix in range(len(self.clusters)):
      if self.clusters[ix].cluname==cluname:
        return self.clusters[ix]
    return None
  def printPart(self):
    print("Partition:{} run:{} status:{}".format(self.pname,self.runn, self.getStatus()))
    if self.errs>0:
      print(self.allerrs)
    else:
      print("Masks:", list(self.masks.keys()))
      print("CLusters:", self.getCluDets()) 
      print("LTGs:", list(self.ltgs.keys()))
  def setStatus(self, newstatus):
      self.status= newstatus
  def getStatus(self):
    if self.status == self.RUNNING:
      st= "R"
    elif self.status == self.LOADED:
      st= "l"
    elif self.status == self.PAUSED:
      st= "P"
    elif self.status == self.LOADEDHW:
      st= "L"
    else:
      st= "?"
    return st
class Partitions(object):
  def __init__(self,ctp=None):
    self.parts= {}   # self.parts[runnumber] -Partition instance
    self.ctp= ctp   # None: started without ctp (e.g. validate.sh)
    self.rsrcs= Resources()
  def addPart(self, pname, runn, dets,alowedltus=None):
    """ create Partition + add it to self.parts only in case LOADED
rc: always Partition object
"""
    #if pname in self.parts:
    #  printError("Partition %s already in, replacing..."%pname)
    np= Partition(pname, runn, dets, alowedltus)
    if np.status==np.LOADED:
      # till now, hw not touched. Check resources, if not enough
      # prevent this partition to be loaded.
      newrsrcs= self.rsrcs + np.getResources()
      if newrsrcs.check(alowedltus):
        self.parts[runn]= np
        self.rsrcs= newrsrcs
      else:
        np.status= None
    else:
      #print("addPart(%s) errors:"%pname, np.errs, '\n', np.allerrs)
      #np= None always return Partition object (even with errors)
      pass
    return np
  def delPart(self, runn):
    if runn in self.parts:
      del self.parts[runn]
    else:
      printError("%d not loaded/started"%runn)
  def getN(self):
    return len(self.parts)
  def getPart(self, runn):
    if runn in self.parts:
      return self.parts[runn]
    else:
      return None
  def listParts(self):
    print("RunN  ", "Name   ", "Status")
    for npn in self.parts:
      np= self.parts[npn]
      st= np.getStatus()
      print(npn, np.pname, st)

if LTUS==None: LTUS= Ltus()

def main():
  # parted.py partname
  if len(sys.argv) < 2 or (len(sys.argv)==2 and sys.argv[1]=="help"):
    print("""
Usage: partdef.py pname[/det1,det2,...] pname[/det1,det2,...] [runn]
pname     -partition definition in CTP3_DBCTP/pd/pname.pd
/det1,det2 -list of detectors (if not present all dets in pname.pd taken) NOT IMPLEMENTED YET
runn      -runn (0 or missing: -> validate partition )

todo:
1.
detectors -filter out given detectors

2. interactive cmdline mode: check more partitions -i.e. should be able to keep a replica
of what is loaded in ctp

""")
    sys.exit()    
  runn= 0
  parts= Partitions()
  for ipar in range(len(sys.argv)):
    dets= ""
    if ipar<1: continue
    par= sys.argv[ipar]
    #if par.isnumeric():  only py3
    if par.isdigit():
      runn= int(par)
      break
    #if len(sys.argv) > 3: dets= " ".join(sys.argv[3:])
    pname= par
    part= parts.addPart(pname, runn, dets)
    #part.printPart()
    if part.status!=part.LOADED:
      print("=== {} not valid".format(pname))
      #print(part.allerrs)
      #sys.exit(8)
    else:
      print("=== {} valid (LOADED)".format(pname))
      #part.printPart()
    print()
  print("=== ctp resources:")
  parts.rsrcs.prt()
  
if __name__ == "__main__":
  main()

